# This fixture creates targets named a.txt through z.txt, and
# distributes them across 8 hashed bin delegations. This uses
# the "classic" (i.e., not succinct) form of hashed bins.

import string

from fixtures.builder import FixtureBuilder
from tuf import roledb

def build():
    builder = FixtureBuilder('HashedBins', { 'use_snapshot_length': True })\
        .publish(with_client=True)

    list_of_targets = []
    for c in list(string.ascii_lowercase):
        name = c + '.txt'
        builder.create_target(name, signing_role=None)
        list_of_targets.append(name)

    # We need at least one key that will sign the targets in the hashed bins.
    public_key, private_key = builder._import_key()

    # Create the hashed bins.
    builder.repository.targets.delegate_hashed_bins(list_of_targets, [public_key], 8)

    # Assign the targets we've created to those hashed bins. TUF determines which
    # target goes in which bin.
    for name in list_of_targets:
        builder.repository.targets.add_target_to_bin(name, 8)

    # Ensure the delegated roles that manage the hashed bins can actually be signed.
    # It's weird, but for some reason this is not done by delegate_hashed_bins().
    for role in builder.repository.targets.get_delegated_rolenames():
        builder.repository.targets(role).load_signing_key(private_key)

    # Make all the delegated roles terminating.
    targets_role_info = roledb.get_roleinfo('targets', 'HashedBins')
    for i in range(len(targets_role_info['delegations']['roles'])):
        targets_role_info['delegations']['roles'][i]['terminating'] = True
    roledb.update_roleinfo('targets', targets_role_info, repository_name='HashedBins')

    # Publish these changes on the server side.
    builder.invalidate()
    builder.publish()



"""
Contains a class to help build fixtures programmatically.
"""

from securesystemslib import formats, signer
from tuf import repository_tool, roledb

import json
import os
import shutil
from dirhash import dirhash


class FixtureBuilder:

    def __init__(self, name, tuf_arguments={ 'use_snapshot_length': False }):
        self.dir = os.path.join(os.path.dirname(__file__), name)

        # The index of the next key pair (in the keys/ directory) to use when initializing
        # a role.
        self._key_index = 0
        # The keychain, containing all public and private keys. The dictionary
        # keys are role names, and each item is a dictionary with 'public' and
        # 'private' members, which are lists of public and private keys.
        self._keys = {}
        # The directory of server-side metadata (and targets).
        self._server_dir = os.path.join(self.dir, 'server')

        # If a directory of server-side metadata already exists, remove it.
        if os.path.isdir(self._server_dir):
            shutil.rmtree(self._server_dir)

        self.repository = repository_tool.create_new_repository(self._server_dir, name, **tuf_arguments)
        self.repository.status()

        # Initialize the basic TUF roles.
        self.add_key('root')
        self.add_key('targets')
        self.add_key('snapshot')
        self.add_key('timestamp')

        self.repository.status()

    def __del__(self):
        # Create a hash for the generated fixture.
        with open(self.dir + "/hash.txt", "w") as hash_file:
            hash_file.write(dirhash(self.dir, 'sha256', ignore=["__init__.py", "client_versions.ini", "hash.txt"]))

    def _role(self, name):
        """Loads a role object for a specific role."""
        try:
            return getattr(self.repository, name)
        except AttributeError:
            return self.repository.targets(name)

    def delegate(self, role_name, paths, parent='targets', path_hash_prefixes=None, terminating=False):
        """Creates a delegated role."""
        self._role(parent).delegate(role_name, [], paths, path_hash_prefixes=path_hash_prefixes, terminating=terminating)
        self.add_key(role_name)
        return self

    def add_key(self, role_name):
        """Loads a key pair from disk and assigns it to a given role."""
        (public_key, private_key) = self._import_key()

        role = self._role(role_name)
        role.add_verification_key(public_key)
        role.load_signing_key(private_key)

        if role_name not in self._keys:
            self._keys[role_name] = {'public': [], 'private': []}

        self._keys[role_name]['public'].append(public_key)
        self._keys[role_name]['private'].append(private_key)

        self._mark_dirty(role_name)

        return self

    def revoke_key(self, role_name, key_index=0):
        """Revokes a key pair from a given role."""
        public_key = self._keys[role_name]['public'].pop(key_index)
        self._role(role_name).remove_verification_key(public_key)
        self._keys[role_name]['private'].pop(key_index)

        self._mark_dirty(role_name)

        return self

    def _mark_dirty(self, role_name):
        """Marks a role as dirty, along with its parent role."""
        self.repository.mark_dirty([role_name])

        if role_name in roledb.TOP_LEVEL_ROLES:
            self.repository.mark_dirty(['root'])
        else:
            self.repository.mark_dirty(['targets'])

    def _import_key(self):
        """Loads a key pair from the keys/ directory."""
        keys_dir = os.path.join(os.path.dirname(__file__), 'keys')
        private_key = os.path.join(keys_dir, str(self._key_index)) + '_key'
        public_key = '{}.pub'.format(private_key)

        self._key_index = self._key_index + 1

        return (
            repository_tool.import_ed25519_publickey_from_file(public_key),
            repository_tool.import_ed25519_privatekey_from_file(private_key, password='pw')
        )

    def invalidate(self):
        """Marks the four top-level TUF roles as dirty."""
        self.repository.mark_dirty(roledb.TOP_LEVEL_ROLES)
        return self

    def add_target(self, filename, signing_role='targets'):
        """Adds an existing target file and signs it."""
        # @todo Just use add_target or add_targets consistently. This is only
        # here while fixtures are being ported to FixtureBuilder, to maintain
        # consistency with previously generated fixtures.
        if signing_role == 'targets':
            self._role('targets').add_targets([filename])
        else:
            self._role(signing_role).add_target(filename)
        self.repository.mark_dirty(['snapshot', 'targets', 'timestamp', signing_role])

        return self

    def create_target(self, filename, contents=None, signing_role='targets'):
        """Creates a signed target file with arbitrary contents."""
        if contents is None:
            contents = 'Contents: ' + filename

        path = os.path.join(self._server_dir, 'targets', filename)
        with open(path, 'w') as f:
            f.write(contents)

        if signing_role is not None:
            self.add_target(filename, signing_role)

        return self

    def publish(self, with_client=False, consistent=True):
        """Writes the TUF metadata to disk."""
        self.repository.writeall(consistent_snapshot=consistent)

        staging_dir = os.path.join(self._server_dir, 'metadata.staged')
        live_dir = os.path.join(self._server_dir, 'metadata')
        shutil.copytree(staging_dir, live_dir, dirs_exist_ok=True)

        if with_client:
            client_dir = os.path.join(self.dir, 'client')
            # If a directory of client-side metadata already exists, remove it.
            if os.path.isdir(client_dir):
                shutil.rmtree(client_dir)

            repository_tool.create_tuf_client_directory(self._server_dir, client_dir)

        return self

    def read(self, filename):
        """Returns the parsed contents of an existing metadata file."""
        path = os.path.join(self._server_dir, 'metadata', filename)

        with open(path, 'r') as f:
            return json.load(f)

    def write(self, filename, data):
        path = os.path.join(self._server_dir, 'metadata', filename)

        with open(path, 'w') as f:
            json.dump(data, f, indent=1, separators=(',', ': '), sort_keys=True)

    def write_signed(self, filename, data, signing_role):
        """Writes arbitrary metadata, signed with a given role's keys, to a file."""
        self.write(filename, {
            'signatures': self._sign(data, signing_role),
            'signed': data
        })

    def _sign(self, data, signing_role):
        """Signs arbitrary data using a given role's keys."""
        signatures = []

        # Encode the data to canonical JSON, which is what we will actually sign.
        data = str.encode(formats.encode_canonical(data))

        # Loop through the signing role's private keys and use each one to sign
        # the canonical JSON representation of the data.
        for key in self._keys[signing_role]['private']:
            signature = signer.SSlibSigner(key).sign(data)
            signatures.append(signature.to_dict())

        return signatures


class ConsistencyVariantFixtureBuilder:

    def __init__(self, name, tuf_arguments={ 'use_snapshot_length': False }):
        self.fixtures = [
            FixtureBuilder(os.path.join(name, 'consistent'), tuf_arguments),
            FixtureBuilder(os.path.join(name, 'inconsistent'), tuf_arguments)
        ]

    def delegate(self, role_name, paths, parent='targets', path_hash_prefixes=None, terminating=False):
        for fixture in self.fixtures:
            fixture.delegate(role_name, paths, parent, path_hash_prefixes, terminating)
        return self

    def add_key(self, role_name):
        for fixture in self.fixtures:
            fixture.add_key(role_name)
        return self

    def revoke_key(self, role_name, key_index=0):
        for fixture in self.fixtures:
            fixture.revoke_key(role_name, key_index)
        return self

    def invalidate(self):
        for fixture in self.fixtures:
            fixture.invalidate()
        return self

    def add_target(self, filename, signing_role='targets'):
        for fixture in self.fixtures:
            fixture.add_target(filename, signing_role)
        return self

    def create_target(self, filename, contents=None, signing_role='targets'):
        for fixture in self.fixtures:
            fixture.create_target(filename, contents, signing_role)
        return self

    def publish(self, with_client=False):
        self.fixtures[0].publish(with_client, consistent=True)
        self.fixtures[1].publish(with_client, consistent=False)
        return self

    def read(self, filename):
        return [
            self.fixtures[0].read(filename),
            self.fixtures[1].read(filename)
        ]

    def write(self, filename, data):
        for fixture in self.fixtures:
            fixture.write(filename, data)

    def write_signed(self, filename, data, signing_role):
        for fixture in self.fixtures:
            fixture.write_signed(filename, data, signing_role)



# For instructions on using this script, please see the README.

from unittest import mock
import shutil
import glob
import os

from fixtures import (
    Simple,
    AttackRollback,
    Delegated,
    HashedBins,
    NestedDelegated,
    NestedDelegatedErrors,
    ThresholdTwo,
    TerminatingDelegation,
    TopLevelTerminating,
    NestedTerminatingNonDelegatingDelegation,
    ThreeLevelDelegation,
    PublishedTwice,
    TargetsLengthNoSnapshotLength
)


@mock.patch('time.time', mock.MagicMock(return_value=1577836800))
def generate_fixtures():
    Simple.build()
    AttackRollback.build()
    Delegated.build()
    HashedBins.build()
    NestedDelegated.build()
    NestedDelegatedErrors.build()
    ThresholdTwo.build()
    TerminatingDelegation.build()
    TopLevelTerminating.build()
    NestedTerminatingNonDelegatingDelegation.build()
    ThreeLevelDelegation.build()
    PublishedTwice.build()
    PublishedTwice.build(rotate_keys='timestamp')
    PublishedTwice.build(rotate_keys='snapshot')
    TargetsLengthNoSnapshotLength.build()


# Remove all previous fixtures.
for f in glob.glob("fixtures/**/client"):
    shutil.rmtree(f)
for f in glob.glob("fixtures/**/server"):
    shutil.rmtree(f)
# Delete hash files to ensure they are generated again.
for f in glob.glob("fixtures/**/hash.txt"):
    os.remove(f)
generate_fixtures()




from fixtures.builder import ConsistencyVariantFixtureBuilder


def build():
    fixture = ConsistencyVariantFixtureBuilder('NestedDelegatedErrors')\
        .create_target('testtarget.txt')\
        .publish(with_client=True)\
        .delegate('unclaimed', ['level_1_*.txt'])\
        .create_target('level_1_target.txt', signing_role='unclaimed')\
        .publish(with_client=True)
    # === Point of No Return ===
    # Past this point, we don't re-export the client. This supports testing the
    # client's own ability to pick up and trust new data from the repository.
    fixture.add_key('targets')\
        .add_key('snapshot')\
        .invalidate()\
        .publish()\
        .revoke_key('targets')\
        .revoke_key('snapshot')\
        .invalidate()\
        .publish()

    # Delegate from level_1_delegation to level_2
    fixture.delegate('level_2', ['level_1_2_*.txt'], parent='unclaimed')\
        .create_target('level_1_2_target.txt', signing_role='level_2')

    # Create a terminating delegation
    fixture.delegate('level_2_terminating', ['level_1_2_terminating_*.txt'], parent='unclaimed', terminating=True)\
        .create_target('level_1_2_terminating_findable.txt', signing_role='level_2_terminating')

    # Create a delegation under non-terminating 'level_2' delegation.
    fixture.delegate('level_3', ['level_1_2_3_*.txt'], parent='level_2')\
        .create_target('level_1_2_3_below_non_terminating_target.txt', signing_role='level_3')

    # Add a delegation below the 'level_2_terminating' role.
    # Delegations from a terminating role are evaluated but delegations after a terminating delegation
    # are not.
    # See NestedDelegatedErrors
    fixture.delegate('level_3_below_terminated', ['level_1_2_terminating_3_*.txt'], parent='level_2_terminating')\
        .create_target('level_1_2_terminating_3_target.txt', signing_role='level_3_below_terminated')

    fixture.publish()

    # Add a target that does not match the path for the delegation.
    fixture.create_target('level_a.txt', signing_role='unclaimed')
    # Add a target that matches the path parent delegation but not the current delegation.
    fixture.create_target('level_1_3_target.txt', signing_role='level_2')
    # Add a target that does not match the delegation's paths.
    fixture.create_target('level_2_unfindable.txt', signing_role='level_2_terminating')

    # Add a delegation after level_2_terminating which will not be evaluated.
    fixture.delegate('level_2_after_terminating_match_terminating_path', ['level_1_2_terminating_plus_1_more_*.txt'], parent='unclaimed')\
        .create_target('level_1_2_terminating_plus_1_more_unfindable.txt', signing_role='level_2_after_terminating_match_terminating_path')\
        .publish()



from fixtures.builder import ConsistencyVariantFixtureBuilder


def build():
    builder = ConsistencyVariantFixtureBuilder('ThresholdTwo')\
        .add_key('timestamp')
    for fixture in builder.fixtures:
        fixture._role('timestamp').threshold = 2
        fixture.repository.mark_dirty(['timestamp'])
    builder.publish(with_client=True)



from fixtures.builder import ConsistencyVariantFixtureBuilder


def build():
    """
    Generates a TUF test fixture that publishes twice -- once on the client,
    and twice on the server -- with targets length information, but no
    snapshot length information.
    """
    ConsistencyVariantFixtureBuilder('TargetsLengthNoSnapshotLength', { 'use_timestamp_length': False, 'use_snapshot_length': True })\
    .publish(with_client=True)\
    .publish()



from fixtures.builder import FixtureBuilder

import os


def build():
    _build(True)
    _build(False)

def _build(consistent):
    if consistent is True:
        suffix = 'consistent'
    else:
        suffix = 'inconsistent'

    name = os.path.join('ThresholdTwoAttack', suffix)

    fixture = FixtureBuilder(name)\
        .add_key('timestamp')
    fixture._role('timestamp').threshold = 2
    fixture.repository.mark_dirty(['timestamp'])
    fixture.publish(with_client=True, consistent=consistent)
    fixture.repository.mark_dirty(['timestamp'])
    fixture.publish(with_client=True, consistent=consistent)

    # By exporting the repo but not the client, this gives us a new revision
    # that's ready to alter. If we alter a version the client is already
    # aware of, it may not pick up this new, altered version.
    fixture.repository.mark_dirty(['timestamp'])
    fixture.publish(consistent=consistent)

    fixture.add_key('timestamp')

    timestamp = fixture.read('timestamp.json')
    timestamp["signatures"][1] = {
        'keyid': fixture._keys['timestamp']['public'][2]['keyid'],
        # This is the SHA-512 hash of the sentence "This is just a random string".
        'sig': 'd1f9ee4f5861ad7b8be61c0c00f3cd4353cee60e70db7d6fbeab81b75e6a5e3871276239caf93d09e9cd406ba764c31abe00e95f2553a3cb543874cb6e7d1545'
    }

    fixture.write('timestamp.json', timestamp)

    # We could also alter the versioned (N.timestamp.json), but the spec
    # considers these as optional, so we can expect this alteration to be
    # sufficient.



from fixtures.builder import ConsistencyVariantFixtureBuilder


def build():
    fixture = ConsistencyVariantFixtureBuilder('Delegated', { 'use_timestamp_length': False })\
        .create_target('testtarget.txt')\
        .publish(with_client=True)\
        .delegate('unclaimed', ['level_1_*.txt'])\
        .create_target('level_1_target.txt', signing_role='unclaimed')\
        .publish(with_client=True)
    # === Point of No Return ===
    # Past this point, we don't re-export the client. This supports testing the
    # client's own ability to pick up and trust new data from the repository.
    fixture.add_key('targets')\
        .add_key('snapshot')\
        .invalidate()\
        .publish()\
        .revoke_key('targets')\
        .revoke_key('snapshot')\
        .invalidate()\
        .publish()



from fixtures.builder import ConsistencyVariantFixtureBuilder


def build():
    ConsistencyVariantFixtureBuilder('Simple', { 'use_snapshot_hashes': True })\
        .create_target('testtarget.txt')\
        .publish(with_client=True)



# You should never need to run this, but this is the code that generated the
# fixed keys for fixture creation.

from tuf import repository_tool as rt


def write_and_import_keypair(filename):
    pathpriv = '{}_key'.format(filename)
    rt.generate_and_write_ed25519_keypair(pathpriv, password='pw')


for i in range(20):
    write_and_import_keypair(i)



# Delegation tree
#
#             Targets
#             /     \
#            a       b
#
# a is the only terminating delegation
#
# Roles should be evaluated in the order:
# Targets > a
#
# Role b should not be evaluated.
from fixtures.builder import ConsistencyVariantFixtureBuilder


def build():
    ConsistencyVariantFixtureBuilder('TopLevelTerminating')\
        .publish(with_client=True)\
        .create_target('targets.txt')\
        .delegate('a', ['*.txt'], terminating=True)\
        .create_target('a.txt', signing_role='a')\
        .delegate('b', ['*.txt'])\
        .create_target('b.txt', signing_role='b')\
        .publish()
